import warnings
import random
import numpy as np


def check_environment_deterministic(env1, env2):
    '''
    env1 and env2 should be seeded environments

    returns a bool: true if env1 and env2 execute the same way
    '''

    # checks deterministic behavior if seed is set
    actions = {agent: space.sample() for agent, space in env1.action_spaces.items()}
    hashes = []
    num_seeds = 2
    envs = [env1, env2]
    for x in range(num_seeds):
        new_env = envs[x]
        cur_hashes = []
        for i in range(3):
            obs = new_env.reset()
            for i in range(x + 1):
                random.randint(0, 1000)
                np.random.normal(size=100)
            cur_hashes.append(hash_obsevation(obs))
            for _ in range(50):
                rew, done, info = new_env.last()
                if done:
                    break
                next_obs = new_env.step(actions[new_env.agent_selection])
                cur_hashes.append(hash_obsevation(next_obs))
                cur_hashes.append(float(rew))

        hashes.append(hash(tuple(cur_hashes)))

    return all(hashes[0] == h for h in hashes)


def hash_obsevation(obs):
    try:
        val = hash(obs.tobytes())
        return val
    except AttributeError:
        try:
            return hash(obs)
        except TypeError:
            warnings.warn("Observation not an int or an Numpy array")
            return 0


def seed_test(env_constructor):
    env1 = env_constructor()
    env2 = env_constructor()
    base_seed = 42
    env1.seed(base_seed)
    env2.seed(base_seed)

    assert check_environment_deterministic(env1, env2), \
         ("The environment gives different results on multiple runs when intialized with the same seed. This is usually a sign that you are using np.random or random modules directly, which uses a global random state.")

    env1.seed(base_seed)
    env2.seed(base_seed+1)
    if check_environment_deterministic(env1, env2):
        warnings.warn("The environment gives same results on multiple runs when intialized by default. By default, environments that take a seed argument should be nondeterministic")



class ArgVar:
    meta = "bob"
    def __init__(self):
        pass

class VarArg(ArgVar):
    meta = "bob"
    def __init__(self):
        self.meta = "joe"

print(VarArg().meta)

exit(0)
import gym
from supersuit.gym_wrappers import color_reduction_v0, frame_stack
from supersuit.vector.gym_rollout import RolloutBuilder,transpose_rollout,split_rollouts_on_dones
from supersuit.vector.gym_rollout import RandomPolicy
from stable_baselines.common.vec_env import DummyVecEnv
def make_gym_env():
    env = gym.make('SpaceInvaders-v0')

    env = frame_stack(color_reduction_v0(env, 'full'), 4)
    return env

num_envs = 4
envs = DummyVecEnv([make_gym_env]*num_envs)
policy = RandomPolicy(envs.action_space, num_envs)
rolloutbuilder = RolloutBuilder(envs)
rolloutbuilder.restart(policy)
n_steps = 5
obs,rew,done,infos = rolloutbuilder.rollout(policy,n_steps)
assert obs.shape[:2] == (n_steps,num_envs)
assert rew.shape == (n_steps,num_envs)
assert done.shape == (n_steps,num_envs)
tobs,trew,tdone,tinfos = transpose_rollout(obs,rew,done,infos)
assert tobs.shape[:2] == (num_envs,n_steps)
assert trew.shape == (num_envs,n_steps)
assert tdone.shape == (num_envs,n_steps)
tobs,trew,tinfos = split_rollouts_on_dones(tobs,trew,tdone,tinfos)
assert len(tobs) == num_envs
# exit(0)
# import pettingzoo.gamma
# from supersuit.aec_wrappers import color_reduction_v0, frame_stack
#
# def make_env():
#     env = pettingzoo.gamma.pistonball_v0.env()
#
#     env = frame_stack(color_reduction_v0(env, 'full'), 4)
#     return env
#
# envs = gym.vector.SyncVectorEnv([make_env]*4)



from supersuit.aec_wrappers import frame_skip
import pettingzoo
from pettingzoo.mpe import simple_push_v0
import numpy as np

env = simple_push_v0.raw_env()
env = frame_skip(env,(2,4))
env.reset()

game_len = 0
for x in range(20000):
    #env.render()
    while True:
        reward, done, info = env.last()

        game_len += 1
        if done:

            print("finished!!!\n\n\n")

            orig_obs = env.reset()

            break

        action = env.action_spaces[env.agent_selection].sample()

        next_obs = env.step(action)
        print(env.env.steps)



from pettingzoo.atari import boxing_v0
from pettingzoo.atari import combat_tank_v0
from pettingzoo.atari import combat_plane_v0
from pettingzoo.atari import double_dunk_v1
from pettingzoo.atari import entombed_competitive_v1
from pettingzoo.atari import entombed_cooperative_v0
from pettingzoo.atari import flag_capture_v0
from pettingzoo.atari import joust_v1
from pettingzoo.atari import ice_hockey_v0
from pettingzoo.atari import maze_craze_v1
from pettingzoo.atari import mario_bros_v1
from pettingzoo.atari import othello_v1
from pettingzoo.atari import pong_basketball_v0
from pettingzoo.atari import pong_classic_v0
from pettingzoo.atari import pong_foozpong_v0
from pettingzoo.atari import pong_quadrapong_v1
from pettingzoo.atari import pong_volleyball_v0
from pettingzoo.atari import space_invaders_v0
from pettingzoo.atari import space_war_v0
from pettingzoo.atari import surround_v0
from pettingzoo.atari import tennis_v1
from pettingzoo.atari import video_checkers_v1
from pettingzoo.atari import wizard_of_wor_v1
from pettingzoo.atari import warlords_v1

from pettingzoo.classic import chess_v0
from pettingzoo.classic import rps_v0
from pettingzoo.classic import rpsls_v0
from pettingzoo.classic import connect_four_v0
from pettingzoo.classic import tictactoe_v0
from pettingzoo.classic import leduc_holdem_v0
from pettingzoo.classic import mahjong_v0
from pettingzoo.classic import texas_holdem_v0
from pettingzoo.classic import texas_holdem_no_limit_v0
from pettingzoo.classic import uno_v0
from pettingzoo.classic import dou_dizhu_v0
from pettingzoo.classic import gin_rummy_v0
from pettingzoo.classic import go_v0
from pettingzoo.classic import hanabi_v0
from pettingzoo.classic import backgammon_v0

from pettingzoo.butterfly import knights_archers_zombies_v2
from pettingzoo.butterfly import pistonball_v0
from pettingzoo.butterfly import cooperative_pong_v1
from pettingzoo.butterfly import prison_v1
from pettingzoo.butterfly import prospector_v1

from pettingzoo.magent import battle_v1
from pettingzoo.magent import adversarial_pursuit_v1
from pettingzoo.magent import gather_v1
from pettingzoo.magent import combined_arms_v1
from pettingzoo.magent import tiger_deer_v1
from pettingzoo.magent import battlefield_v1

from pettingzoo.mpe import simple_adversary_v0
from pettingzoo.mpe import simple_crypto_v0
from pettingzoo.mpe import simple_push_v0
from pettingzoo.mpe import simple_reference_v0
from pettingzoo.mpe import simple_speaker_listener_v0
from pettingzoo.mpe import simple_spread_v0
from pettingzoo.mpe import simple_tag_v0
from pettingzoo.mpe import simple_world_comm_v0
from pettingzoo.mpe import simple_v0

from pettingzoo.sisl import pursuit_v1
from pettingzoo.sisl import waterworld_v1
from pettingzoo.sisl import multiwalker_v2

all_prefixes = ["atari", "classic", "butterfly", "magent", "mpe", "sisl"]

manual_environments = {
    "butterfly/knights_archers_zombies",
    "butterfly/pistonball",
    "butterfly/cooperative_pong",
    "butterfly/prison",
    "butterfly/prospector",
    "sisl/pursuit"
}

all_environments = {
    "atari/boxing": boxing_v0,
    "atari/combat_tank": combat_tank_v0,
    "atari/combat_plane": combat_plane_v0,
    "atari/double_dunk": double_dunk_v1,
    "atari/entombed_cooperative": entombed_cooperative_v0,
    "atari/entombed_competitive": entombed_competitive_v1,
    "atari/flag_capture": flag_capture_v0,
    "atari/joust": joust_v1,
    "atari/ice_hockey": ice_hockey_v0,
    "atari/maze_craze": maze_craze_v1,
    "atari/mario_bros": mario_bros_v1,
    "atari/othello": othello_v1,
    "atari/pong_classic": pong_classic_v0,
    "atari/pong_basketball": pong_basketball_v0,
    "atari/pong_foozpong": pong_foozpong_v0,
    "atari/pong_quadrapong": pong_quadrapong_v1,
    "atari/pong_volleyball": pong_volleyball_v0,
    "atari/space_invaders": space_invaders_v0,
    "atari/space_war": space_war_v0,
    "atari/surround": surround_v0,
    "atari/tennis": tennis_v1,
    "atari/video_checkers": video_checkers_v1,
    "atari/wizard_of_wor": wizard_of_wor_v1,
    "atari/warlords": warlords_v1,

    "classic/chess": chess_v0,
    "classic/rps": rps_v0,
    "classic/rpsls": rpsls_v0,
    "classic/connect_four": connect_four_v0,
    "classic/tictactoe": tictactoe_v0,
    "classic/leduc_holdem": leduc_holdem_v0,
    "classic/mahjong": mahjong_v0,
    "classic/texas_holdem": texas_holdem_v0,
    "classic/texas_holdem_no_limit": texas_holdem_no_limit_v0,
    "classic/uno": uno_v0,
    "classic/dou_dizhu": dou_dizhu_v0,
    "classic/gin_rummy": gin_rummy_v0,
    "classic/go": go_v0,
    "classic/hanabi": hanabi_v0,
    "classic/backgammon": backgammon_v0,

    "butterfly/knights_archers_zombies": knights_archers_zombies_v2,
    "butterfly/pistonball": pistonball_v0,
    "butterfly/cooperative_pong": cooperative_pong_v1,
    "butterfly/prison": prison_v1,
    "butterfly/prospector": prospector_v1,

    "magent/adversarial_pursuit": adversarial_pursuit_v1,
    "magent/battle": battle_v1,
    "magent/battlefield": battlefield_v1,
    "magent/combined_arms": combined_arms_v1,
    "magent/gather": gather_v1,
    "magent/tiger_deer": tiger_deer_v1,

    "mpe/simple_adversary": simple_adversary_v0,
    "mpe/simple_crypto": simple_crypto_v0,
    "mpe/simple_push": simple_push_v0,
    "mpe/simple_reference": simple_reference_v0,
    "mpe/simple_speaker_listener": simple_speaker_listener_v0,
    "mpe/simple_spread": simple_spread_v0,
    "mpe/simple_tag": simple_tag_v0,
    "mpe/simple_world_comm": simple_world_comm_v0,
    "mpe/simple": simple_v0,

    "sisl/multiwalker": multiwalker_v2,
    "sisl/waterworld": waterworld_v1,
    "sisl/pursuit": pursuit_v1,
}



from gym.spaces import Box, Discrete
from .dummy_aec_env import DummyEnv
import numpy as np
from supersuit import frame_stack_v0,reshape_v0,observation_lambda_v0,action_lambda_v0,pad_action_space_v0,pad_observations_v0,dtype_v0
import supersuit
import pytest


base_obs = {"a{}".format(idx): np.zeros([8,8,3],dtype=np.float32) + np.arange(3) + idx for idx in range(2)}
base_obs_space = {"a{}".format(idx): Box(low=np.float32(0.),high=np.float32(10.),shape=[8,8,3]) for idx in range(2)}
base_act_spaces = {"a{}".format(idx): Discrete(5) for idx in range(2)}

def test_frame_stack():
    base_obs_space = {"a{}".format(idx): Box(low=np.float32(0.),high=np.float32(10.),shape=[2,3]) for idx in range(2)}
    base_obs = {"a{}".format(idx): np.zeros([2,3]) + np.arange(3) + idx for idx in range(2)}
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = frame_stack_v0(base_env, 4)
    obs = env.reset()
    assert obs.shape == (2,3,4)
    first_obs = env.step(2)
    print(first_obs[:,:,-1])
    assert np.all(np.equal(first_obs[:,:,-1],base_obs["a1"]))
    assert np.all(np.equal(first_obs[:,:,:-1],0))

    base_obs = {"a{}".format(idx): idx+3 for idx in range(2)}
    base_env = DummyEnv(base_obs, base_act_spaces, base_act_spaces)
    env = frame_stack_v0(base_env, 4)
    obs = env.reset()
    assert env.observation_spaces[env.agent_selection].n == 5**4
    first_obs = env.step(2)
    assert first_obs == 4
    second_obs = env.step(2)
    assert second_obs == 3+3*5
    for x in range(100):
        nth_obs = env.step(2)
    assert nth_obs == ((3*5+3)*5+3)*5+3

def test_frame_skip():
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = supersuit.frame_skip_v0(base_env, 3)
    env.reset()
    for i in range(10):
        env.step(0)

def test_agent_indicator():
    let = ["a","a","b"]
    base_obs = {"{}_{}".format(let[idx],idx): np.zeros([2,3]) for idx in range(3)}
    base_obs_space = {"{}_{}".format(let[idx],idx): Box(low=np.float32(0.),high=np.float32(10.),shape=[2,3]) for idx in range(3)}
    base_act_spaces = {"{}_{}".format(let[idx],idx): Discrete(5) for idx in range(3)}

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = supersuit.agent_indicator_v0(base_env,type_only=True)
    obs = env.reset()
    assert obs.shape == (2,3,3)
    assert env.observation_spaces["a_0"].shape == (2,3,3)
    first_obs = env.step(2)

    env = supersuit.agent_indicator_v0(base_env,type_only=False)
    obs = env.reset()
    assert obs.shape == (2,3,4)
    assert env.observation_spaces["a_0"].shape == (2,3,4)
    first_obs = env.step(2)

def test_reshape():
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = reshape_v0(base_env, (64, 3))
    obs = env.reset()
    assert obs.shape == (64,3)
    first_obs = env.step(5)
    assert np.all(np.equal(first_obs,base_obs["a1"].reshape([64,3])))


def new_continuous_dummy():

    base_obs = {"a_{}".format(idx): (np.zeros([8,8,3],dtype=np.float32) + np.arange(3) + idx).astype(np.float32) for idx in range(2)}
    base_obs_space = {"a_{}".format(idx): Box(low=np.float32(0.),high=np.float32(10.),shape=[8,8,3]) for idx in range(2)}
    base_act_spaces = {"a_{}".format(idx): Box(low=np.float32(0.),high=np.float32(10.),shape=[3]) for idx in range(2)}

    return  DummyEnv(base_obs, base_obs_space, base_act_spaces)

def new_dummy():

    base_obs = {"a_{}".format(idx): (np.zeros([8,8,3],dtype=np.float32) + np.arange(3) + idx).astype(np.float32) for idx in range(2)}
    base_obs_space = {"a_{}".format(idx): Box(low=np.float32(0.),high=np.float32(10.),shape=[8,8,3]) for idx in range(2)}
    base_act_spaces = {"a_{}".format(idx): Discrete(5) for idx in range(2)}

    return  DummyEnv(base_obs, base_obs_space, base_act_spaces)

wrappers = [
    supersuit.color_reduction_v0(new_dummy(),"R"),
    supersuit.resize_v0(dtype_v0(new_dummy(),np.uint8),x_size=5,y_size=10),
    supersuit.resize_v0(dtype_v0(new_dummy(),np.uint8),x_size=5,y_size=10,linear_interp=True),
    supersuit.dtype_v0(new_dummy(),np.int32),
    supersuit.flatten_v0(new_dummy()),
    supersuit.reshape_v0(new_dummy(),(64,3)),
    supersuit.normalize_obs_v0(new_dummy(),env_min=-1,env_max=5.),
    supersuit.frame_stack_v0(new_dummy(),8),
    supersuit.pad_observations_v0(new_dummy()),
    supersuit.pad_action_space_v0(new_dummy()),
    supersuit.agent_indicator_v0(new_dummy(),True),
    supersuit.agent_indicator_v0(new_dummy(),False),
    #supersuit.normalize_reward(new_dummy()),
    supersuit.reward_lambda_v0(new_dummy(), lambda x:x/10),
    supersuit.clip_reward_v0(new_dummy()),
    supersuit.clip_actions_v0(new_continuous_dummy()),
    supersuit.frame_skip_v0(new_dummy(), 4),
    supersuit.sticky_actions_v0(new_dummy(), 0.75),
    supersuit.delay_observations_v0(new_dummy(), 3),
]
@pytest.mark.parametrize("env", wrappers)
def test_basic_wrappers(env):
    env.seed(5)
    obs = env.reset()
    act_space = env.action_spaces[env.agent_selection]
    obs_space = env.observation_spaces[env.agent_selection]
    first_obs = env.observe("a_0")
    assert obs_space.contains(first_obs)
    assert first_obs.dtype == obs_space.dtype
    env.step(act_space.sample())
    for i in env.agent_iter(20):
        act_space = env.action_spaces[env.agent_selection]
        env.step(act_space.sample())


def test_rew_lambda():
    env = supersuit.reward_lambda_v0(new_dummy(), lambda x:x/10)
    env.reset()
    assert env.rewards[env.agent_selection] == 1./10

def test_lambda():
    def add1(obs):
        return obs+1
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = observation_lambda_v0(base_env, add1)
    obs0 = env.reset()
    assert int(obs0[0][0][0]) == 1
    env = observation_lambda_v0(env, add1)
    obs0 = env.reset()
    assert int(obs0[0][0][0]) == 2

    def tile_obs(obs):
        shape_size = len(obs.shape)
        tile_shape = [1]*shape_size
        tile_shape[0] *= 2
        return np.tile(obs,tile_shape)

    env = observation_lambda_v0(env, tile_obs)
    obs0 = env.reset()
    assert env.observation_spaces[env.agent_selection].shape == (16,8,3)
    def change_shape_fn(obs_space):
        return Box(low=0,high=1,shape=(32,8,3))
    env = observation_lambda_v0(env, tile_obs)
    obs0 = env.reset()
    assert env.observation_spaces[env.agent_selection].shape == (32,8,3)
    assert obs0.shape == (32,8,3)

def test_action_lambda():
    def inc1(x,space):
        return (x + 1)
    def change_space_fn(space):
        return Discrete(space.n+1)
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = action_lambda_v0(base_env, inc1, change_space_fn)
    env.reset()
    env.step(5)

    def one_hot(x,n):
        v = np.zeros(n)
        v[x] = 1
        return v

    act_spaces = {"a{}".format(idx): Box(low=0,high=1,shape=(15,)) for idx in range(2)}
    base_env = DummyEnv(base_obs, base_obs_space, act_spaces)
    env = action_lambda_v0(base_env,
        lambda action, act_space : one_hot(action, act_space.shape[0]),
        lambda act_space : Discrete(act_space.shape[0]))

    env.reset()
    env.step(2)

def test_dehomogenize():
    base_act_spaces = {"a{}".format(idx): Discrete(5+idx) for idx in range(2)}

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = pad_action_space_v0(base_env)
    env.reset()
    assert all([s.n == 6 for s in env.action_spaces.values()])
    env.step(5)



from supersuit.basic_transforms.resize import check_param,change_observation
from gym.spaces import Box
import numpy as np
import pytest

test_shape = (6,4,3)
high_val = (np.ones(test_shape) + np.arange(4).reshape(1,4,1))
test_obs_space = Box(low=high_val-1,high=high_val,dtype=np.uint8)
test_val = high_val-0.5

def test_param_check():
    check_param(test_obs_space, (2,2,False))
    with pytest.raises(AssertionError):
        check_param(test_obs_space, (-2,2,2))
    with pytest.raises(AssertionError):
        check_param(test_obs_space, (2.,2,2))

def test_change_observation():
    cur_val = test_val
    #print(cur_val)
    cur_val = cur_val.astype(np.uint8)
    #print(cur_val)
    new_obs = change_observation(cur_val,test_obs_space,(3,2,False))
    new_obs = change_observation(cur_val,test_obs_space,(3,2,True))
    test_obs = np.array(
    [[[0.6666667, 0.6666667, 0.6666667],
      [2.       , 2.       , 2.       ],
      [3.3333335, 3.3333335, 3.3333335]],

     [[0.6666667, 0.6666667, 0.6666667],
      [2.       , 2.       , 2.       ],
      [3.3333335, 3.3333335, 3.3333335]]]
    ).astype(np.uint8)
    assert new_obs.dtype == np.uint8
    assert np.all(np.equal(new_obs,test_obs))

    test_shape = (6,4)
    high_val = np.ones(test_shape).astype(np.float64)
    obs_spae = Box(low=high_val-1,high=high_val)
    new_obs = change_observation(high_val-0.5,obs_spae,(3,2,False))
    assert new_obs.shape == (2,3)
    assert new_obs.dtype == np.float64

    test_shape = (6,5,4)
    high_val = np.ones(test_shape).astype(np.uint8)
    obs_spae = Box(low=high_val-1,high=high_val,dtype=np.uint8)
    new_obs = change_observation(high_val,obs_spae,(5,2,False))
    assert new_obs.shape == (2,5,4)
    assert new_obs.dtype == np.uint8



'''
Base environment definitions

See docs/api.md for api documentation
'''


class AECEnv:
    def __init__(self):
        pass

    def step(self, action, observe=True):
        raise NotImplementedError

    def reset(self, observe=True):
        raise NotImplementedError

    def seed(self, seed=None):
        raise NotImplementedError

    def observe(self, agent):
        raise NotImplementedError

    def render(self, mode='human'):
        raise NotImplementedError

    def close(self):
        pass


class ParallelEnv:
    def reset(self):
        raise NotImplementedError

    def seed(self, seed=None):
        raise NotImplementedError

    def step(self, actions):
        raise NotImplementedError

    def render(self, mode="human"):
        raise NotImplementedError

    def close(self):
        pass



import numpy as np
import copy
from gym.spaces import Box, Discrete
from gym import spaces
import warnings
from pettingzoo import AECEnv

from .env_logger import EnvLogger


class BaseWrapper(AECEnv):
    '''
    Creates a wrapper around `env` parameter. Extend this class
    to create a useful wrapper.
    '''
    metadata = {'render.modes': ['human']}

    def __init__(self, env):
        super().__init__()
        self.env = env

        self.num_agents = self.env.num_agents
        self.agents = self.env.agents
        self.observation_spaces = self.env.observation_spaces
        self.action_spaces = self.env.action_spaces

        # we don't want these defined as we don't want them used before they are gotten

        # self.agent_selection = self.env.agent_selection

        # self.rewards = self.env.rewards
        # self.dones = self.env.dones

        # we don't want to care one way or the other whether environments have an infos or not before reset
        try:
            self.infos = self.env.infos
        except AttributeError:
            pass

    def seed(self, seed=None):
        self.env.seed(seed)

    def close(self):
        self.env.close()

    def render(self, mode='human'):
        self.env.render(mode)

    def reset(self, observe=True):
        observation = self.env.reset(observe)

        self.agent_selection = self.env.agent_selection
        self.rewards = self.env.rewards
        self.dones = self.env.dones
        self.infos = self.env.infos

        return observation

    def observe(self, agent):
        return self.env.observe(agent)

    def step(self, action, observe=True):
        next_obs = self.env.step(action, observe=observe)

        self.agent_selection = self.env.agent_selection
        self.rewards = self.env.rewards
        self.dones = self.env.dones
        self.infos = self.env.infos

        return next_obs


class AgentIterWrapper(BaseWrapper):
    def __init__(self, env):
        super().__init__(env)
        self._has_updated = False
        self._is_iterating = False
        self._agent_idxs = {agent: i for i, agent in enumerate(env.agents)}
        # self._was_dones = {agent: False for agent in self.agents}

    def reset(self, observe=True):
        self._has_updated = True
        self._final_rewards = {agent: 0 for agent in self.agents}
        self.old_observation = None
        obs = super().reset(observe)
        self._was_dones = {agent: done for agent, done in self.dones.items()}
        return obs

    def last(self):
        agent = self.agent_selection
        return self.rewards[agent], self.dones[agent], self.infos[agent]

    def step(self, action, observe=True):
        self._has_updated = True

        cur_agent = self.agent_selection
        if self.agent_selection != self.env.agent_selection:
            self._was_dones[cur_agent] = True
            self.agent_selection = self.env.agent_selection
        else:
            self._was_dones[cur_agent] = self.env.dones[cur_agent]
            super().step(action, False)

        for agent in self.agents:
            if self.dones[agent] and not self._final_rewards[agent]:
                self._final_rewards[agent] = self.rewards[agent]

        for agent in self.agents:
            if self.dones[agent] and not self._was_dones[agent]:
                self.agent_selection = agent
                self.rewards[agent] = self._final_rewards[agent]
                break

        return super().observe(self.agent_selection) if observe else None

    def agent_iter(self, max_agent_iter=2**63):
        return AECIterable(self, max_agent_iter)


class AECIterable:
    def __init__(self, env, max_agent_iter):
        self.env = env
        self.max_agent_iter = max_agent_iter

    def __iter__(self):
        return AECOrderEnforcingIterator(self.env, self.max_agent_iter)


class AECIterator:
    def __init__(self, env, max_agent_iter):
        self.env = env
        self.iters_til_term = max_agent_iter
        self.env._is_iterating = True

    def __next__(self):
        if self.env._was_dones[self.env.agent_selection] or self.iters_til_term <= 0:
            raise StopIteration
        self.iters_til_term -= 1
        return self.env.agent_selection


class AECOrderEnforcingIterator(AECIterator):
    def __next__(self):
        agent = super().__next__()
        assert self.env._has_updated, "need to call step() or reset() in a loop over `agent_iter`!"
        self.env._has_updated = False
        return agent


class TerminateIllegalWrapper(BaseWrapper):
    '''
    this wrapper terminates the game with the current player losing
    in case of illegal values

    parameters:
        - illegal_reward: number that is the value of the player making an illegal move.
    '''
    def __init__(self, env, illegal_reward):
        super().__init__(env)
        self._illegal_value = illegal_reward

    def step(self, action, observe=True):
        current_agent = self.agent_selection
        assert 'legal_moves' in self.infos[current_agent], "Illegal moves must always be defined to use the TerminateIllegalWrapper"
        if action not in self.infos[current_agent]['legal_moves']:
            EnvLogger.warn_on_illegal_move()
            self.dones = {d: True for d in self.dones}
            for info in self.infos.values():
                info['legal_moves'] = []
            self.rewards = {d: 0 for d in self.dones}
            self.rewards[current_agent] = self._illegal_value
        else:
            return super().step(action, observe)


class NanNoOpWrapper(BaseWrapper):
    '''
    this wrapper expects there to be a no_op_action parameter which
    is the action to take in cases when nothing should be done.
    '''
    def __init__(self, env, no_op_action, no_op_policy):
        super().__init__(env)
        self._no_op_action = no_op_action
        self._no_op_policy = no_op_policy

    def step(self, action, observe=True):
        if np.isnan(action).any():
            EnvLogger.warn_action_is_NaN(self._no_op_policy)
            action = self._no_op_action
        return super().step(action, observe)


class NanZerosWrapper(BaseWrapper):
    '''
    this wrapper warns and executes a zeros action when nothing should be done.
    Only for Box action spaces.
    '''
    def __init__(self, env):
        super().__init__(env)
        assert all(isinstance(space, Box) for space in self.action_spaces.values()), "should only use NanZerosWrapper for Box spaces. Use NanNoOpWrapper for discrete spaces"

    def step(self, action, observe=True):
        if np.isnan(action).any():
            EnvLogger.warn_action_is_NaN("taking the all zeros action")
            action = np.zeros_like(action)
        return super().step(action, observe)


class NaNRandomWrapper(BaseWrapper):
    '''
    this wrapper takes a random action
    '''
    def __init__(self, env):
        super().__init__(env)
        assert all(isinstance(space, Discrete) for space in env.action_spaces.values()), "action space should be discrete for NaNRandomWrapper"
        SEED = 0x33bb9cc9
        self.np_random = np.random.RandomState(SEED)

    def step(self, action, observe=True):
        if np.isnan(action).any():
            cur_info = self.infos[self.agent_selection]
            if 'legal_moves' in cur_info:
                backup_policy = "taking a random legal action"
                EnvLogger.warn_action_is_NaN(backup_policy)
                action = self.np_random.choice(cur_info['legal_moves'])
            else:
                backup_policy = "taking a random action"
                EnvLogger.warn_action_is_NaN(backup_policy)
                act_space = self.action_spaces[self.agent_selection]
                action = self.np_random.choice(act_space.n)

        return super().step(action, observe)


class AssertOutOfBoundsWrapper(BaseWrapper):
    '''
    this wrapper crashes for out of bounds actions
    Should be used for Discrete spaces
    '''
    def __init__(self, env):
        super().__init__(env)
        assert all(isinstance(space, Discrete) for space in self.action_spaces.values()), "should only use AssertOutOfBoundsWrapper for Discrete spaces"

    def step(self, action, observe=True):
        assert self.action_spaces[self.agent_selection].contains(action), "action is not in action space"
        return super().step(action, observe)


class ClipOutOfBoundsWrapper(BaseWrapper):
    '''
    this wrapper crops out of bounds actions for Box spaces
    '''
    def __init__(self, env):
        super().__init__(env)
        assert all(isinstance(space, Box) for space in self.action_spaces.values()), "should only use ClipOutOfBoundsWrapper for Box spaces"

    def step(self, action, observe=True):
        space = self.action_spaces[self.agent_selection]
        if not space.contains(action):
            assert space.shape == action.shape, "action should have shape {}, has shape {}".format(space.shape, action.shape)

            EnvLogger.warn_action_out_of_bound(action=action, action_space=space, backup_policy="clipping to space")
            action = np.clip(action, space.low, space.high)

        return super().step(action, observe)


class OrderEnforcingWrapper(AgentIterWrapper):
    '''
    check all orders:

    * error on getting rewards, dones, infos, agent_selection before reset
    * error on calling step, observe before reset
    * warn on calling close before render or reset
    * warn on calling step after environment is done
    '''
    def __init__(self, env):
        self._has_reset = False
        self._has_rendered = False
        super().__init__(env)

    def __getattr__(self, value):
        '''
        raises an error message when data is gotten from the env
        which should only be gotten after reset
        '''
        if value == "agent_order":
            raise AttributeError("agent_order has been removed from the API. Please consider using agent_iter instead.")
        elif value in {"rewards", "dones", "infos", "agent_selection"}:
            raise AttributeError("{} cannot be accessed before reset".format(value))
        else:
            raise AttributeError("'{}' object has no attribute '{}'".format(type(self).__name__, value))

    def seed(self, seed=None):
        self._has_reset = False
        super().seed(seed)

    def render(self, mode='human'):
        if not self._has_reset:
            EnvLogger.error_render_before_reset()
        self._has_rendered = True
        super().render(mode)

    def close(self):
        super().close()
        if not self._has_rendered:
            EnvLogger.warn_close_unrendered_env()
        if not self._has_reset:
            EnvLogger.warn_close_before_reset()

        self._has_rendered = False
        self._has_reset = False

    def step(self, action, observe=True):
        if not self._has_reset:
            EnvLogger.error_step_before_reset()
        elif self._was_dones[self.agent_selection]:
            EnvLogger.warn_step_after_done()
            self.dones = {agent: True for agent in self.dones}
            self.rewards = {agent: 0 for agent in self.rewards}
            return super().observe(self.agent_selection) if observe else None
        else:
            return super().step(action, observe)

    def observe(self, agent):
        if not self._has_reset:
            EnvLogger.error_observe_before_reset()
        return super().observe(agent)

    def reset(self, observe=True):
        self._has_reset = True
        return super().reset(observe)



from pettingzoo.utils.depreciated_module import DepreciatedModule

multiwalker_v0 = DepreciatedModule("multiwalker", "v0", "v2")
multiwalker_v0 = DepreciatedModule("multiwalker", "v0", "v2")
pursuit_v0 = DepreciatedModule("pursuit", "v0", "v1")
waterworld_v0 = DepreciatedModule("waterworld", "v0", "v1")



from ._mpe_utils.simple_env import SimpleEnv, make_env
from .scenarios.simple_reference import Scenario
from pettingzoo.utils.to_parallel import parallel_wrapper_fn


class raw_env(SimpleEnv):

    def __init__(self, local_ratio=0.5, max_frames=100):
        assert 0. <= local_ratio <= 1., "local_ratio is a proportion. Must be between 0 and 1."
        scenario = Scenario()
        world = scenario.make_world()
        super().__init__(scenario, world, max_frames, local_ratio)


env = make_env(raw_env)
parallel_env = parallel_wrapper_fn(env)
